#include <stdio.h>
#include <stdlib.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>

#define WIN     1   /* $BF~NO2hA|(B   */
#define WOUT    2   /* $B=PNO2hA|(B   */
#define WWORK   3   /* $B%o!<%/2hA|(B */
#define WKEY    4   /* $B%-!<2hA|(B   */

#define HIGH    255 /* uchar$B7?$N%O%$%l%Y%k(B */
#define LOW     0   /* uchar$B7?$N%m!<%l%Y%k(B */

#define GRAY    1   /* $B%b%N%/%m%b!<%I(B */
#define COLOR   2   /* $B%+%i!<%b!<%I(B */
#define GLEVEL  128 /* $B%Q%l%C%H$r;HMQ$7$F%b%N%/%m2hA|$rI=<($9$k;~$N3,D4?t(B  */
#define CLEVELR 5   /* $B%Q%l%C%H$r;HMQ$7$F%+%i!<2hA|$rI=<($9$k;~$N3,D4?t(B(R) */
#define CLEVELG 6   /* $B%Q%l%C%H$r;HMQ$7$F%+%i!<2hA|$rI=<($9$k;~$N3,D4?t(B(G) */
#define CLEVELB 5   /* $B%Q%l%C%H$r;HMQ$7$F%+%i!<2hA|$rI=<($9$k;~$N3,D4?t(B(B) */

typedef struct _ximage {   /* $B2hA|MQ9=B$BN(B */
  Window win;
  XImage *im;
  Pixmap pix;
} XIMAGE;

Display *d;            /* $BI=<(2hLLMQ%]%$%s%?(B             */
Visual  *vis;          /* $BI=<(2hLL$N>pJs(B                 */
GC      gc;            /* $BIA2h4D6-(B                       */
int     depth;         /* $B2hAG$"$?$j$N%S%C%H?t(B           */
int     pal_mode;      /* $B%Q%l%C%H%b!<%I(B                 */
int     p_table[256];  /* $B%Q%l%C%HHV9f=q$-9~$_MQ%F!<%V%k(B */
GC      pen_gc;
unsigned long pen_color;

XIMAGE  ximage_in;     /* $BF~NO2hA|$N9=B$BN(B   */
XIMAGE  ximage_out;    /* $B=PNO2hA|$N9=B$BN(B   */
XIMAGE  ximage_work;   /* $B%o!<%/2hA|$N9=B$BN(B */
XIMAGE  ximage_key;    /* $B%-!<2hA|$N9=B$BN(B   */

/* $B%W%m%H%?%$%W@k8@(B */
void InitDisplay(int,int,int);
void InitVisual(void);
void InitPal(void);
void InitWindows(int,int,int);
void InitWindow(XIMAGE *,int,int,int,int);
void DisplayImage(unsigned char *,int,int,int);
void ColorDisplay(XIMAGE, unsigned char *,unsigned char *,unsigned char *,int,int,char *);
void SetPenColor(char *,int);
unsigned long mycolor(Display *, char *);
void DrawCircle(int,int,int,int);
void DrawPoint(int,int,int);

/*--- InitDisplay --- $B2hLLI=<($N=i4|2=(B($B%b%N%/%m(B) -------------------------------
     xsize: $B2hA|$N2#J}8~$NBg$-$5(B
     ysize: $B2hA|$N=DJ}8~$NBg$-$5(B
     wn: $B@8@.$9$k%&%$%s%I%&$N?t(B(1$B!A(B4$B$^$G(B)
------------------------------------------------------------------------------*/
void InitDisplay(int xsize, int ysize, int wn)
{
  InitVisual();
  if (vis->class == PseudoColor && depth == 8) {
    printf("--Warning: 24bit color mode is recommended.\n");
    pal_mode = GRAY;
    InitPal();
  }
  InitWindows(xsize,ysize,wn);
}

/*--- InitVisual --- $B%+%i!<2hLL$N@_Dj(B ---------------------------------------*/
void InitVisual(void)
{
  int n;
  unsigned long vmask;
  XVisualInfo *xvinfo, vtemp;

  d = XOpenDisplay(NULL);
  if (d == NULL) {
    printf("Cannot open display.\n");
    exit(0);
  }
  vtemp.class = TrueColor;
  vtemp.depth = 24;
  xvinfo = XGetVisualInfo(d, VisualClassMask | VisualDepthMask, &vtemp, &n);
  if (n) {
    vis = xvinfo->visual;
    depth = xvinfo->depth;
  }
  else {
    vis = DefaultVisual(d,0);
    depth = DefaultDepth(d,0);
  }
}

/*--- InitPal --- $B%Q%l%C%H$N=i4|2=(B - ---------------------------------------*/
void InitPal(void)
{
  Colormap cmap;
  Status   result;
  XColor   xcol;
  int      i, npal;

  cmap = DefaultColormap(d,0);
  if (pal_mode == GRAY) {
    for (i=0; i<GLEVEL; i++) {
      xcol.red   = (i*HIGH/(GLEVEL-1) << 8) & 0xff00;
      xcol.green = (i*HIGH/(GLEVEL-1) << 8) & 0xff00;
      xcol.blue  = (i*HIGH/(GLEVEL-1) << 8) & 0xff00;
      result = XAllocColor(d, cmap, &xcol);
      if (result)
	p_table[i] = xcol.pixel;
      else {
	printf("--Palette not enough.\n");
	exit(0);
      }
    }
  }
  else {
    npal = CLEVELR*CLEVELG*CLEVELB;
    for (i=0; i<npal; i++) {
      xcol.red   = (((i/CLEVELG/CLEVELB) % CLEVELR) 
		    * HIGH / (CLEVELR-1) << 8) & 0xff00;
      xcol.green = (((i/CLEVELB) % CLEVELG) 
		    * HIGH / (CLEVELG-1) << 8) & 0xff00;
      xcol.blue  = (( i % CLEVELB) 
		    * HIGH / (CLEVELB-1) << 8) & 0xff00;
      result = XAllocColor(d, cmap, &xcol);
      if (result)
	p_table[i] = xcol.pixel;
      else {
	printf("--Palette not enough..\n");
	exit(0);
      }
    }
  }
}

/*--- InitWindows --- $BA4%&%#%s%I$r=i4|2=$9$k(B -----------------------------------
     xsize: $B2hA|$N2#J}8~$NBg$-$5(B
     ysize: $B2hA|$N=DJ}8~$NBg$-$5(B
------------------------------------------------------------------------------*/
void InitWindows(int xsize, int ysize, int wn)
{
  unsigned char *image;
  int posi_x,posi_y;

  image = (unsigned char *)calloc(xsize*ysize,1);

  if (image == NULL) {
    printf("--Memory allocation error !\n");
    exit(0);
  }
  
  posi_x = 10;
  posi_y = 10;
  
  if(wn==1 || wn==2 || wn==3 || wn==4){
    InitWindow(&ximage_in, xsize, ysize, posi_x, posi_y);
    XMapWindow(d, ximage_in.win);
    DisplayImage(image,xsize,ysize,WIN);
  }
  if(wn==2 || wn==3 || wn==4){
    InitWindow(&ximage_out, xsize, ysize, posi_x+xsize, posi_y);
    XMapWindow(d, ximage_out.win);
    DisplayImage(image,xsize,ysize,WOUT);
  }
  if(wn==3 || wn==4){
    InitWindow(&ximage_work, xsize, ysize, posi_x, posi_y+ysize);
    XMapWindow(d, ximage_work.win);
    DisplayImage(image,xsize,ysize,WWORK);
  }
  if(wn==4){
    InitWindow(&ximage_key, xsize, ysize, posi_x+xsize, posi_y+ysize);
    XMapWindow(d, ximage_key.win);
    DisplayImage(image,xsize,ysize,WKEY);
  }
  free(image);
}

/*--- InitWindow --- $B%&%#%s%I$r=i4|2=$9$k(B -------------------------------------
     ximage: $B2hA|$N9=B$BN(B
     xsize : $B2hA|$N2#J}8~$NBg$-$5(B
     ysize : $B2hA|$N=DJ}8~$NBg$-$5(B
     posi_x: $B%&%#%s%I$N2#J}8~I=<(0LCV(B
     posi_y: $B%&%#%s%I$N=DJ}8~I=<(0LCV(B
-----------------------------------------------------------------------------*/
void InitWindow(XIMAGE *ximage,int xsize,int ysize,int posi_x,int posi_y)
{
  unsigned long vmask = 0;
  XSetWindowAttributes attr;
  XWindowAttributes watr;
  Colormap colormap;
  
  attr.background_pixel = 0;
  attr.border_pixel = 0;
  attr.background_pixmap = 0;
  attr.border_pixmap = 0;

  colormap = XCreateColormap(d, RootWindow(d,0), vis, AllocNone);
  attr.colormap = colormap;

  vmask = CWBackPixel | CWBorderPixel | CWColormap ;

  ximage->win = XCreateWindow(d, RootWindow(d,0),posi_x,posi_y,
			      xsize,ysize,0,depth,InputOutput,
			      vis,vmask,&attr);
  gc = XCreateGC(d, ximage->win, 0, 0);

  if (vis->class == PseudoColor && depth == 8) {
    ximage->im = XCreateImage(d, vis, 8, ZPixmap, 0, 0,
			      xsize, ysize, 8, 0);
    (ximage->im)->data = (char *)calloc(1, xsize*ysize);
    if ((ximage->im)->data == NULL) {
      printf("--Memory allocation error !!\n");
      exit(0);
    }
    ximage->pix = XCreatePixmap(d, ximage->win, xsize, ysize, 8);
  }
  else if (vis->class == TrueColor || vis->class == DirectColor) {
    ximage->im = XCreateImage(d, vis, depth, ZPixmap, 0, 0,
			      xsize, ysize, 32, 0);
    (ximage->im)->data = (char *)calloc(1, xsize*ysize*4);
    if ((ximage->im)->data == NULL) {
      printf("--Memory allocation error !!\n");
      exit(0);
    }
    ximage->pix = XCreatePixmap(d, ximage->win, xsize, ysize, depth);
  }
  else {
    printf("--Not supported X-window mode !\n");
    exit(0);
  }
  XPutImage(d, ximage->pix, gc, ximage->im, 0, 0, 0, 0, xsize, ysize);
  XSetWindowBackgroundPixmap(d, ximage->win, ximage->pix);  
  XFlush(d);
}

/*--- DisplayImage --- $B2hLL$NI=<((B ---------------------------------------------
     image: $B2hA|%G!<%?(B
     xsize: $B2hA|$N2#J}8~$NBg$-$5(B
     ysize: $B2hA|$N=DJ}8~$NBg$-$5(B
     position: $BI=<(%&%#%s%I(B
-----------------------------------------------------------------------------*/
void DisplayImage(unsigned char *image,int xsize,int ysize, int position)
{
  switch (position) {
    case WIN:
      ColorDisplay(ximage_in, image, image, image, xsize, ysize, "1");
      break;
    case WOUT:
      ColorDisplay(ximage_out, image, image, image, xsize, ysize, "2");
      break;
    case WWORK:
      ColorDisplay(ximage_work, image, image, image, xsize, ysize, "3");
      break;
    case WKEY:
      ColorDisplay(ximage_key, image, image, image, xsize, ysize, "4");
      break;
    default:
      printf("There is no such campas window !!\n");
      break;
  }
}

/*--- ColorDisplay --- $B%&%#%s%I$K%+%i!<2hA|$rI=<($9$k(B -------------------------
     ximage: $B2hA|$N9=B$BN(B
     imager: $B2hA|%G!<%?(B(R)
     imageg: $B2hA|%G!<%?(B(G)
     imageb: $B2hA|%G!<%?(B(B)
     xsize : $B2hA|$N2#J}8~$NBg$-$5(B
     ysize : $B2hA|$N=DJ}8~$NBg$-$5(B     
     wtitle: $B%&%#%s%I$N%?%$%H%k(B
-----------------------------------------------------------------------------*/
void ColorDisplay(XIMAGE ximage, unsigned char *imager,
		  unsigned char *imageg, unsigned char *imageb,
		  int xsize, int ysize, char *wtitle)
{
  int           *pd;
  unsigned char *pdc;
  int           i,j;
  unsigned char *psr0,*psg0,*psb0;
  unsigned char *psr,*psg,*psb;

  XStoreName(d, ximage.win, wtitle);
  if (vis->class == PseudoColor && depth == 8) {
    pdc = (unsigned char *)ximage.im->data;
    if (pal_mode == GRAY) {
      psg = (unsigned char *)imageg;
      for(i=0; i<xsize*ysize; i++) {
	j = *psg++ * (GLEVEL-1) / HIGH;
	*psg++ = p_table[j];
      }
    }
    else {
      printf("Not suport!!\n");
      exit(0);
    }
  }
  
  if (vis->class == TrueColor || vis->class == DirectColor) {
    pd = (int *)ximage.im->data;
    psg = (unsigned char *)imageg;
    if (vis->blue_mask == 255) {
      psr = (unsigned char *)imager;
      psb = (unsigned char *)imageb;
    }
    else {
      psb = (unsigned char *)imager;
      psr = (unsigned char *)imageb;
    }
    for (i=0; i<xsize*ysize; i++) {
      *pd++ = ((int)*psb++ & 0xff)
	+ ((int)*psg++ << 8 & 0xff00)
	+ ((int)*psr++ << 16 & 0xff0000);
    }
  }
  XPutImage(d, ximage.pix, gc, ximage.im, 0, 0, 0, 0, xsize, ysize);
  XCopyArea(d, ximage.pix, ximage.win, gc, 0, 0, xsize, ysize, 0, 0);
  XRaiseWindow(d, ximage.win);
  XFlush(d);
}

  /* $BDI2C(B by Lee 2003.3.19 */
/*   SetPenColor("White",1);
     DrawLine(x1,y1,x2,y2,1); */
void SetPenColor(char *color, int position)
{
  XSetWindowAttributes att;
  att.backing_store=Always;
  switch (position) {
    case WIN:
      pen_color = mycolor(d,color);
      pen_gc =XCreateGC(d,ximage_in.win,0,0);   
      XSetForeground(d,pen_gc,pen_color);
      XChangeWindowAttributes(d,ximage_in.win,CWBackingStore,&att);
      break;
    case WOUT:
      pen_color = mycolor(d,color);
      pen_gc =XCreateGC(d,ximage_out.win,0,0);   
      XSetForeground(d,pen_gc,pen_color);
      XChangeWindowAttributes(d,ximage_out.win,CWBackingStore,&att);     
      break;
    case WWORK:
      pen_color = mycolor(d,color);
      pen_gc =XCreateGC(d,ximage_work.win,0,0);   
      XSetForeground(d,pen_gc,pen_color);
      XChangeWindowAttributes(d,ximage_work.win,CWBackingStore,&att);
      break;
    case WKEY:
      pen_color = mycolor(d,color);
      pen_gc =XCreateGC(d,ximage_key.win,0,0);   
      XSetForeground(d,pen_gc,pen_color);
      XChangeWindowAttributes(d,ximage_key.win,CWBackingStore,&att);           
      break;
    default:
      printf("There is no such campas window !!\n");
      break; 
  }    
}

unsigned long mycolor(Display *display, char *color)
{
  Colormap cmap;
  XColor c0, c1;
  XSetWindowAttributes att;

  cmap = DefaultColormap (display,0);
  XAllocNamedColor (display, cmap, color, &c1, &c0);
  return(c1.pixel);
}

void DrawCircle(int xposi,int yposi,int rad,int position)
{
  switch (position) {
    case WIN:
      XDrawArc(d,ximage_in.win,pen_gc,xposi,yposi,rad,rad,64,360*64);
      XFlush(d);
      break;
    case WOUT:
      XDrawArc(d,ximage_out.win,pen_gc,xposi,yposi,rad,rad,64,360*64);
      XFlush(d);
      break;
    case WWORK:
      XDrawArc(d,ximage_work.win,pen_gc,xposi,yposi,rad,rad,64,360*64);
      XFlush(d);
      break;
    case WKEY:
      XDrawArc(d,ximage_key.win,pen_gc,xposi,yposi,rad,rad,64,360*64);
      XFlush(d);
      break;
    default:
      printf("There is no such campas window !!\n");
      break; 
  }
}  

void DrawPoint(int xposi,int yposi,int position)
{
  switch (position) {
    case WIN:
      XDrawPoint(d,ximage_in.win,pen_gc,xposi,yposi);
      XFlush(d);
      break;
    case WOUT:
      XDrawPoint(d,ximage_out.win,pen_gc,xposi,yposi);
      XFlush(d);
      break;
    case WWORK:
      XDrawPoint(d,ximage_work.win,pen_gc,xposi,yposi);
      XFlush(d);
      break;
    case WKEY:
      XDrawPoint(d,ximage_key.win,pen_gc,xposi,yposi);
      XFlush(d);
      break;
    default:
      printf("There is no such campas window !!\n");
      break; 
  }
}  
void DrawLine(int xposi,int yposi,int xposi2,int yposi2,int position)
{
  switch (position) {
    case WIN:
      XDrawLine(d,ximage_in.win,pen_gc,xposi,yposi,xposi2,yposi2);
      XFlush(d);
      break;
    case WOUT:
      XDrawLine(d,ximage_out.win,pen_gc,xposi,yposi,xposi2,yposi2);
      XFlush(d);
      break;
    case WWORK:
      XDrawLine(d,ximage_work.win,pen_gc,xposi,yposi,xposi2,yposi2);
      XFlush(d);
      break;
    case WKEY:
      XDrawLine(d,ximage_key.win,pen_gc,xposi,yposi,xposi2,yposi2);
      XFlush(d);
      break;
    default:
      printf("There is no such campas window !!\n");
      break; 
  }
}  

